module TypeScript.Writers exposing
    ( autoGeneratedFileWarning
    , declareModule
    , initFn
    , interface
    , namespace
    , portFns
    )

{-| Templating system for building TypeScript declaration files. A Writer
encapsulates the layout rules for a block of text, end defines how nested child
Writers are composed.
-}

import String exposing (join)
import String.Interpolate exposing (interpolate)
import Util.String exposing (indented)
import Writer exposing (Writer, writer)


{-| Writer for a comment warning that the file is auto-generated.
-}
autoGeneratedFileWarning : Writer
autoGeneratedFileWarning =
    let
        content =
            [ "// WARNING: Do not manually modify this file. It was generated using:"
            , "// https://github.com/mulias/elm-tigershark"
            , "// Type definitions for using Elm programs in TypeScript"
            , ""
            ]
                |> join "\n"
    in
    writer { format = always content, children = [] }


declareModule : String -> List Writer -> Writer
declareModule name children =
    let
        format contents =
            interpolate "declare module {0} {\n{1}\n}"
                [ name, indented 2 (join "\n" contents) ]
    in
    writer
        { format = format, children = children }


{-| Writer which wraps `children` in a namespace declaration. For example:

    ```
    /** A doc comment */
    export namespace Foo {
        // children
    }
    ```

-}
namespace : { docs : Maybe String, export : Bool, name : String } -> List Writer -> Writer
namespace { docs, export, name } children =
    let
        docStr =
            docs
                |> Maybe.map (\str -> str ++ "\n")
                |> Maybe.withDefault ""

        exportStr =
            if export then
                "export "

            else
                ""

        format contents =
            interpolate "{0}{1}namespace {2} {\n{3}\n}"
                [ docStr, exportStr, name, indented 2 (join "\n" contents) ]
    in
    writer
        { format = format, children = children }


{-| Writer which wraps `children` in an interface declaration. For example:

    ```
    export interface Bar {
      // children
    }
    ```

-}
interface : { export : Bool, name : String } -> List Writer -> Writer
interface { export, name } children =
    let
        exportStr =
            if export then
                "export "

            else
                ""

        format contents =
            interpolate "{0}interface {1} {\n{2}\n}"
                [ exportStr, name, indented 2 (join "\n" contents) ]
    in
    writer
        { format = format, children = children }


{-| Writer for a declaration for each port function. For example:

    ```
    ports: {
      portFoo: {
        send(data: string): void;
      }
    }
    ```

When the ports list is empty, an empty object is used:

    ```
    ports: {}
    ```

-}
portFns : List { name : String, body : String } -> Writer
portFns portDefs =
    let
        format contents =
            interpolate "ports: {{0}{1}{2}};"
                (if List.isEmpty contents then
                    [ "", "", "" ]

                 else
                    [ "\n", indented 2 (join "\n" contents), "\n" ]
                )
    in
    writer { format = format, children = List.map portDeclaration portDefs }


{-| Writer for individual ports inside the ports object.
-}
portDeclaration : { name : String, body : String } -> Writer
portDeclaration { name, body } =
    let
        content =
            interpolate "{0}: {\n{1};\n};" [ name, indented 2 body ]
    in
    writer { format = always content, children = [] }


{-| Writer for the init function used to mount Elm 0.19 programs. The
`moduleName` is needed for the namespaced function return type, and the flags
are used as the second argument to the function, if the program accepts flags.
-}
initFn : { moduleName : String, flags : Maybe String } -> Writer
initFn { moduleName, flags } =
    let
        flagsStr =
            flags
                |> Maybe.map (\str -> "flags: " ++ str ++ ";")
                |> Maybe.withDefault ""

        nextLine =
            flags |> Maybe.map (always "\n") |> Maybe.withDefault ""

        template =
            """export function init(options: {
  node?: HTMLElement | null;{0}{1}
}): Elm.{2}.App;"""

        content =
            interpolate template [ nextLine, indented 2 flagsStr, moduleName ]
    in
    writer { format = always content, children = [] }
